/*!
 * @file
 * @brief implements Join_IResultSetAggregator
 *
 * @author GertG
 * @ingroup Join
 *
 * @par last changed by:
 * <br>
 * $Author: d024980 $ $DateTime: 2005/02/18 17:24:39 $
 *
 * @sa Join_IResultSetAggregator.hpp
 */
/*

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


*/
#include "Join/Join_IResultSetAggregator.hpp"
#include "SAPDBCommon/SAPDB_RangeCode.hpp"
#include "SAPDBCommon/SAPDB_MemCopyMove.hpp"
#include "hak07.h"
#include "hbd07.h"
#include "hkb721.h"
#include "hbd01.h"
#include "hgg01_3.h"
#include "hta01.h"
#include "hsp51.h"

/* ******************** PUBLIC MEMBERS ********************* */

/*!
* destructor
*/
Join_IResultSetAggregator::~Join_IResultSetAggregator()
{
    SAPDBTRACE_METHOD_DEBUG( "Join_IResultSetAggregator::dtor", Join_Trace, 5 );
    
    if ( ! m_DistinctFileRoot.IsEmpty() )
    {
        // destroy distinct files
        b01prefix_destroy_files(m_acv.TransContext(), 
                m_DistinctFileId.fileName_gg00(), TEMP_PREFIX_MXGG00 + 1 );
    }
}

/* ******************* PROTECTED MEMBERS ******************* */

/* ******************** PRIVATE MEMBERS ******************** */
/*
 * analyses result record and detect ROWNO/ROWNUM positions
 */
tgg00_BasisError Join_IResultSetAggregator::initialize() 
{
    SAPDBTRACE_METHOD_DEBUG( "Join_IResultSetAggregator::initialize", Join_Trace, 1 );

    tgg00_BasisError _b_err = e_ok;
    SAPDB_UInt4 _i = 1;
    SAPDB_UInt4 _func_pos = (*m_RecordDesc)[ 0 ].epos() - 1;
    SAPDB_UInt2 _func_cnt = 0, _out_pos;
    SAPDB_UInt4 _disfile_cnt = 0;
   
    // check for LASTFUNCTION
    if ( !( st_func == (*m_RecordDesc)[ _func_pos - 1 ].etype() && 
          op_f_none == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ))
        _b_err = e_stack_type_illegal;
    
    while (( _func_pos - (*m_RecordDesc)[ _func_pos - 1 ].epos() > 1) 
            && ( e_ok == _b_err ))
    {
        _func_pos -= (*m_RecordDesc)[ _func_pos - 1 ].epos();
        // loop over LASTFUNCTION
        while (( st_func == (*m_RecordDesc)[ _func_pos - 1 ].etype() ) 
                && ( op_f_none == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ))
            _func_pos -= (*m_RecordDesc)[ _func_pos - 1 ].epos();
        SAPDBTRACE_WRITELN( Join_Trace, 3, "func pos: " << _func_pos );
        if ( st_func == (*m_RecordDesc)[ _func_pos - 1 ].etype() )
        {
            if ( op_f_check_null != (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
            {
                ++_func_cnt; 
                if  (
                    (op_f_avg         == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_avg      == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_stddev       == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_stddev   == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_variance     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_variance == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
                    &&
                    (
                    (1 == (*m_RecordDesc)[ _func_pos - 1 ].ecol_tab()[0]) ||
                    /* selrec.selr_union OR       check for union has to be included */
                    ((ttfnUserResult_egg00 != m_ResultTree.fileTfnTemp_gg00()) &&
                    ( ttfnInto_egg00       != m_ResultTree.fileTfnTemp_gg00()))
                    )
                    )
                {
                    m_CallPostProcessing = sfh_all;
                }
                else
                {
                    if (op_f_stddev       == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                        op_f_dis_stddev   == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                        op_f_variance     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                        op_f_dis_variance == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
                    { 
                        m_CallPostProcessing = sfh_stddev_variance;
                    }
                }
                if ( op_f_dis_avg     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_sum      == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_count    == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_stddev       == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_stddev   == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_variance     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||
                    op_f_dis_variance == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
                {
                    tgg00_FileId _htree( m_DistinctFileId );
                    
                    _htree.fileTempCnt_gg00()   = ++_disfile_cnt;
                    _htree.fileRootCheck_gg00() = ROOT_CHECK_GG00;
                    b01tcreate_file( m_acv.TransContext(), _htree );
                    _b_err = m_acv.TransContext().trError_gg00;
                    if ( m_DistinctFileRoot.IsFull() ) 
                        m_DistinctFileRoot.Reserve( 2*m_DistinctFileRoot.GetSize() );
                    m_DistinctFileRoot.InsertEnd( Join_RootPair(_htree.fileRoot_gg00(), _htree.fileRootCheck_gg00()) );
                }
                _out_pos = k721out_entry( (tgg00_StackList*) m_RecordDesc.GetPtr(), _func_pos );
                if ( op_o_output_hold == (*m_RecordDesc)[ _out_pos - 1 ].eop_out() &&
                    op_o_output_later == (*m_RecordDesc)[ _out_pos - 1 + 1 ].eop_out() )
                    m_LaterOutputPos = _out_pos + 1;
           }
        }
        else
            _b_err = e_stack_type_illegal;
    }
    SAPDBTRACE_WRITELN( Join_Trace, 3, "m_CallPostProcessing: " << m_CallPostProcessing );

    return _b_err;
}

/* ********************************************************* */
/*!
 *
 */
void Join_IResultSetAggregator::post_processing()
{
    SAPDBTRACE_METHOD_DEBUG( "Join_IResultSetAggregator::post_processing", Join_Trace, 3 );

    tgg00_BasisError _b_err = e_ok;
    bool _repl_needed;
    SAPDB_UInt4 _cnt_repl = 0;
    SAPDB_UInt4 _func_pos, _out_pos;
    SAPDB_UInt4 _disfile_cnt = 0;
    SAPDB_UInt4 _res_cnt = 0;
    tgg00_Rec _buffer;
    Container_Vector<SAPDB_Int4>::Iterator _iter;

    this->init_result_scan();
    do {
        SAPDBTRACE_IF( Join_Trace, 5, SAPDB_MemFillNoCheck( &_buffer, ':', sizeof( _buffer ) ) );
        _func_pos = (*m_RecordDesc)[ 0 ].epos() - 1;
        _b_err = this->fetch_next_result( _buffer );
        
        if ( m_acv.TransContext().trRteCommPtr_gg00->to_cancel )
            _b_err = e_cancelled;
        if ( e_ok == _b_err )
        {
            ++_res_cnt;
            _repl_needed = false;
            
            if ( !m_RecordCounterPositions.IsEmpty() )
            {
                tsp00_NumError _ret;
        
                // put ROWNO
                for ( _iter = m_RecordCounterPositions.Begin(); _iter != m_RecordCounterPositions.End(); ++_iter) 
                {
                    _buffer.buf()[ *_iter - 1 ] = csp_defined_byte;
                    s41plint( _buffer.buf(), *_iter + 1,
                          csp_resnum_deflen, 0, _res_cnt, _ret );
                }
                _repl_needed = true;
            }

            while (( _func_pos - (*m_RecordDesc)[ _func_pos - 1 ].epos() > 1) 
                    && ( e_ok == _b_err ))
            {
                _func_pos -= (*m_RecordDesc)[ _func_pos - 1 ].epos();
                // loop over LASTFUNCTION
                while (( st_func == (*m_RecordDesc)[ _func_pos - 1 ].etype() ) 
                        && ( op_f_none == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ))
                    _func_pos -= (*m_RecordDesc)[ _func_pos - 1 ].epos();
                _out_pos = k721out_entry( (tgg00_StackList*) m_RecordDesc.GetPtr(), _func_pos );
                if ( st_func == (*m_RecordDesc)[ _func_pos - 1 ].etype() )
                {
                    if  (
                        (
                        (op_f_avg       == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() || 
                        op_f_dis_avg    == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                        op_f_sum        == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                        op_f_dis_sum    == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
                        &&
                        (sfh_all == m_CallPostProcessing)
                        )                                                              ||
                        op_f_stddev     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                        op_f_dis_stddev == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                        op_f_variance   == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                        op_f_dis_variance == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() 
                        )
                    {
                        if ( csp_undef_byte != _buffer.recBuf_gg00()[ (*m_RecordDesc)[ _out_pos - 1 ].epos() - 1 ] && 
                            csp_undef_byte == _buffer.recBuf_gg00()[ (*m_RecordDesc)[ _out_pos - 1 ].epos() - 1 + 1 ] )
                        {
                            // check for s51add() errors 
                            SAPDB_UInt4 _i = (*m_RecordDesc)[ _out_pos - 1 ].epos() + 2;
                            bool _overflow = true;
                            do {
                                if ( 0x99 != _buffer.recBuf_gg00()[ _i - 1 ] )
                                    _overflow = false;
                                ++_i;
                            } while (!( _i == (*m_RecordDesc)[ _out_pos - 1 ].epos() + 10 || !_overflow ));
                            if ( _overflow )
                            {
                                _buffer.recBuf_gg00()[ (*m_RecordDesc)[ _out_pos - 1 + 1 ].epos() - 1 ] = csp_oflw_byte;
                                _repl_needed = true;
                            }
                        }
                    }
                    if ( op_f_avg        == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() || 
                         op_f_dis_avg    == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                         op_f_stddev     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                         op_f_dis_stddev == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                         op_f_variance   == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                         op_f_dis_variance == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
                    {
                        if ( csp_undef_byte != _buffer.recBuf_gg00()[ (*m_RecordDesc)[ _out_pos - 1 ].epos() - 1 ] && 
                             csp_oflw_byte  != _buffer.recBuf_gg00()[ (*m_RecordDesc)[ _out_pos - 1 ].epos() - 1 ] )
                        {
                            tsp00_NumError _numerr;
                            SAPDB_Int4 _length;
                            SAPDBTRACE_WRITELN( Join_Trace, 3, "s51div " );
#ifdef SAPDB_SLOW
                            SAPDBTRACE_IF ( Join_Trace, 5, t01buf( td_always, &_buffer, 
                                    (*m_RecordDesc)[ _out_pos - 1 ].epos() + 1, 
                                    (*m_RecordDesc)[ _out_pos - 1 ].epos() + 1 + NUMBER_MXGG04 - 1 ));
                            SAPDBTRACE_IF ( Join_Trace, 5, t01buf( td_always, &_buffer, 
                                    (*m_RecordDesc)[ _out_pos - 1 ].epos() + 1 + NUMBER_MXGG04, 
                                    (*m_RecordDesc)[ _out_pos - 1 ].epos() + 1 + NUMBER_MXGG04 + RESNUM_MXSP00 - 1 ));
#endif
                            s51div( &_buffer,
                                  (*m_RecordDesc)[ _out_pos - 1 ].epos() + 1, 
                                  NUMBER_MXGG04 - 1,
                                  &_buffer, 
                                  (*m_RecordDesc)[ _out_pos - 1 ].epos() + NUMBER_MXGG04 + 1,
                                  RESNUM_MXSP00 - 1,
                                  &_buffer, 
                                  (*m_RecordDesc)[ _out_pos - 1 ].epos() + 1,
                                  csp_fixed, csp_float_frac, _length, _numerr );
                            _repl_needed = true;
                            if ( op_f_stddev     == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                                 op_f_dis_stddev == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                                 op_f_variance   == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() ||    
                                 op_f_dis_variance == (*m_RecordDesc)[ _func_pos - 1 ].eop_func() )
                            {
                                tgg00_FileId _htree( m_DistinctFileId );
                                tgg00_Rec _buf1, _buf2;
                                SAPDBTRACE_IF( Join_Trace, 5, SAPDB_MemFillNoCheck( &_buf1, ':', sizeof( _buf1 ) ) );
                                SAPDBTRACE_IF( Join_Trace, 5, SAPDB_MemFillNoCheck( &_buf2, ':', sizeof( _buf2 ) ) );
                                _htree.fileRoot_gg00() = m_DistinctFileRoot[ _disfile_cnt ].m_Root;
                                _htree.fileRootCheck_gg00() = m_DistinctFileRoot[ _disfile_cnt ].m_RootCheck;
                                _htree.fileTempCnt_gg00() = ++_disfile_cnt;
                                
                                k721stddev( &m_acv.TransContext(), _htree, 
                                        _buffer, _buf1, &_buf2, 
                                        (*m_RecordDesc)[ _func_pos - 1 ].eop_func(),
                                        (*m_RecordDesc)[ _out_pos - 1 ].epos() );
                            }
                        }
                    }
                }
            }
            // call later output
            if ( m_LaterOutputPos > 0 )
            {
                SAPDBTRACE_WRITELN( Join_Trace, 3, "put 'later out': " 
                        << (*m_RecordDesc)[ m_LaterOutputPos - 1 - 1 ].epos() 
                        << "[" << (*m_RecordDesc)[ m_LaterOutputPos - 1 ].elen_var() << "]" 
                        << " --> " << (*m_RecordDesc)[ m_LaterOutputPos - 1 ].epos() );
                SAPDB_RangeMove( __FILE__, 8,
                        sizeof( _buffer ), sizeof( _buffer ),
                        &_buffer, 
                        (*m_RecordDesc)[ m_LaterOutputPos - 1 - 1 ].epos(),
                        &_buffer, 
                        (*m_RecordDesc)[ m_LaterOutputPos - 1 ].epos(),
                        (*m_RecordDesc)[ m_LaterOutputPos - 1 ].elen_var(), 
                        m_acv.TransContext().trError_gg00 );
            }

            SAPDBTRACE_IF( Join_Trace, 5, 
            t01buf( td_always, &_buffer, 1, _buffer.recLen_gg00() <= 300 ? _buffer.recLen_gg00() : 300 ));
            if ( _repl_needed && ( e_ok == (_b_err = update_result( _buffer)) )) ++_cnt_repl;
        }
    } while ( e_ok == _b_err && e_ok == m_acv.TransContext().trError_gg00 );
    
    if ( e_no_next_record == m_acv.TransContext().trError_gg00 )
        m_acv.TransContext().trError_gg00 = e_ok;
    b73cmds_count( iins_sorts_rows_ins, _cnt_repl );
}

